/*====================================+=====================================+
! File CChannel.cpp                   ! Copyright (C) 2016-2020 Remi PASCAL !
+-------------------------------------+-------------------------------------+
! This file is part of matelefree.                                          !
! It is free software: you can redistribute it and/or modify it under the   !
! terms of the GNU General Public License as published by the Free Software !
! Foundation, either version 3 of the License, or any later version.        !
! This application is distributed in the hope that it will be useful, but   !
! WITHOUT ANY WARRANTY; without even the implied warranty of                !
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General !
! Public License for more details.                                          !
! You should have received a copy of the GNU General Public License along   !
! with this application. If not, see <http://www.gnu.org/licenses/>.        !
+-------+-------------------------------------------------------------------+
! Notes !                                                                   !
+-------+                                                                   !
!                                                                           !
! Twitch parsing code bellow is based on :                                  !
! Twitch.tv playlist parser v0.0.4 by Stefan Sundin                         !
!                                                                           !
+==========================================================================*/



/*-------------------------------------------------------------------------*/
#include <wx/clipbrd.h>
#include <wx/filename.h>
#include <wx/textctrl.h>
#include <wx/stattext.h>
#include <wx/button.h>
#include <wx/dir.h>
#include <wx/sstream.h>
#include <wx/txtstrm.h>
#include <wx/regex.h>
#include <wx/imaglist.h>
#include <curl/curl.h>
#include "matelefree.h"
#include "app_info.h"
#include "CApp.h"
#include "CVlcInterface.h"
#include "common/sc_download.h"
#include "CFrame.h"
#include "CPlaylist.h"
/*-------------------------------------------------------------------------*/



/*-------------------------------------------------------------------------*/
wxArrayString str_stream::m_st_as_quality ;

/*-------------------------------------------------------------------------*/
wxString str_stream::get_vlc_url() const
{
   /*----------------------------------------------------------------------*/
   wxString s_url = m_s_url ;

   /*--( Adding the VLC options )------------------------------------------*/
   for( const wxString &s_option : m_vs_option )
   {  s_url += " :" + s_option ; }

   /*----------------------------------------------------------------------*/
   return s_url ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void str_stream::st_init()
{
   /*-----------------------------------------------------------------------+
   ! The different stream qualities as defined by the "Freebox"             !
   ! This will be present in prefs and used to sort channel streams         !
   +-----------------------------------------------------------------------*/
   m_st_as_quality.Add( "auto" ) ;
   m_st_as_quality.Add( wxString::FromUTF8( u8"bas d\u00e9bit" ) ) ;
   m_st_as_quality.Add( "standard" ) ;
   m_st_as_quality.Add( "HD" ) ;
   m_st_as_quality.Add( "TNT" ) ;
   m_st_as_quality.Add( "TNT HD" ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CChannel::CChannel( const wxString &s_pl, const wxString &s_name, int i_num )
                  : m_s_pl_source( s_pl ), m_i_num( i_num )
{
   /*----------------------------------------------------------------------*/
   set_short_name( s_name ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CChannel::CChannel( e_pl_known pl, const wxString &s_name, int i_num )
                  : m_s_pl_source( wxGetApp().get_pk_name( pl ) ),
                    m_pl_known( pl ),
                    m_i_num( i_num )
{
   /*----------------------------------------------------------------------*/
   set_short_name( s_name ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannel::get_vlcver_min() const
{  return CVlcInterface::m_st_co_i_version_min ; }

/*-------------------------------------------------------------------------*/
wxString CChannel::build_name( const wxString &s_short_name ) const
{
   /*----------------------------------------------------------------------*/
   return s_short_name + " [" + m_s_pl_source + ']' ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CChannel::set_pl_source( const wxString &s_pl_source )
{
   /*----------------------------------------------------------------------*/
   m_s_pl_source = s_pl_source ;
   /*--( The source is part of the "long" name )---------------------------*/
   m_s_name = build_name( m_s_short_name ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CChannel::set_short_name( const wxString &s_name )
{
   /*----------------------------------------------------------------------*/
   m_s_short_name = s_name ;
   m_s_name_find = mtf::prepare_string( m_s_short_name ) ;
   m_s_name = build_name( m_s_short_name ) ;
   /*--( The logo is based on the channel short name )---------------------*/
   init_logo_fn() ;
   /*--( The logo has to be recreated )------------------------------------*/
   m_boo_logo_ok = false ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CChannel::get_pl_text_short() const
{
   /*----------------------------------------------------------------------*/
   if( !is_num_set() ) { return get_short_name() ; }
   /*----------------------------------------------------------------------*/
   return wxString::Format( "%d - %s", m_i_num, get_short_name() ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CChannel::get_pl_text_long() const
{
   /*----------------------------------------------------------------------*/
   if( !is_num_set() ) { return get_name() ; }
   /*----------------------------------------------------------------------*/
   return wxString::Format( "%d - %s", m_i_num, get_name() ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CChannel::get_title()
{
   /*----------------------------------------------------------------------*/
   wxString s_title = get_pl_text_long() ;
   /*--( Stream info is given only if more than one are available )--------*/
   if( m_vec_stream.size() > 1 && !m_s_stream_name.empty() )
   {  s_title += " (" + m_s_stream_name + ')' ; }
   /*----------------------------------------------------------------------*/
   return s_title ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CChannel::init_logo_fn()
{
   /*-----------------------------------------------------------------------+
   ! How to find the channel logo name ?                                    !
   ! The search is based on the channel name.                               !
   !                                                                        !
   ! How is structured a channel name ? : channel_base_name (stream name)   !
   ! For example, the following channels will share the same logo :         !
   ! Arte (HD)                                                              !
   ! Arte (francais)                                                        !
   !                                                                        !
   ! The text between '(' and ')', the stream name, must be at the end.     !
   ! The logo search is done from the start of the name to the " (" of      !
   ! the stream name.                                                       !
   !                                                                        !
   ! To avoid issues with the OS naming restrictions the search is done     !
   ! after making the following changes :                                   !
   ! - Unaccent                                                             !
   ! - Lowercase                                                            !
   ! - Replace "special" characters with '-'                                !
   +-----------------------------------------------------------------------*/
   sc::t_map_strstr_cit cit_name
                      = wxGetApp().m_map_image_fn.find( get_name_find() ) ;
   /*----------------------------------------------------------------------*/
   if( cit_name == wxGetApp().m_map_image_fn.end() )
   {  m_s_logo_fn.clear() ; }
   else
   {  /*--( No swap because different channels can have same names )-------*/
      m_s_logo_fn = cit_name->second ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*--( Add "en option" channel logo detail )--------------------------------*/
void CChannel::draw_en_option_overlay( wxBitmap &bmp )
{
   /*--( The "bmp" parameter is not supposed to have transparent points )--*/
   wxMemoryDC dc( bmp ) ;
   wxColour   colour_gold( 255, 215, 0 ) ;
   wxColour   colour_goldenrod( 218, 165, 32 ) ;
   wxBrush    brush_gold( colour_gold ) ;
   wxPen      pen_gold( colour_goldenrod ) ;
   /*--( Draw upper left triangle )----------------------------------------*/
   dc.SetBrush( brush_gold ) ;
   dc.SetPen( pen_gold ) ;
   wxPoint a_pt[] = { { 0, 0 }, { 15, 0 }, { 0, 15 } } ;
   dc.DrawPolygon( WXSIZEOF( a_pt ), a_pt ) ;
   /*--( Draw frame )------------------------------------------------------*/
   dc.SetBrush( *wxTRANSPARENT_BRUSH ) ;
   dc.DrawRectangle( { 0, 0 }, bmp.GetSize() ) ;
   dc.SelectObject( wxNullBitmap ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxBitmap CChannel::build_logo()
{
   /*----------------------------------------------------------------------*/
   wxString s_name = get_short_name() ;
   const size_t co_sz_name = s_name.size() ;
   wxUint8 a_b_rgb[ 3 ] ;
   /*--( Background colour computed based on text )------------------------*/
   a_b_rgb[ 0 ] = a_b_rgb[ 1 ] = a_b_rgb[ 2 ] = 0xAA ;
   for( size_t sz_num = 0 ; sz_num < co_sz_name ; ++sz_num )
   {  a_b_rgb[ sz_num % 3 ] ^= ( s_name[ sz_num ].GetValue() << 4 ) ; }
   wxColour color_background( a_b_rgb[ 0 ], a_b_rgb[ 1 ], a_b_rgb[ 2 ] ) ;
   /*--( Foreground colour computed based on the background one )----------*/
   double do_brightness
    = a_b_rgb[ 0 ] * 0.299 + a_b_rgb[ 1 ] * 0.587 + a_b_rgb[ 2 ] * 0.114 ;
   wxColour color_foreground( do_brightness > 160 ? *wxBLACK : *wxWHITE ) ;
   /*----------------------------------------------------------------------*/
   wxBitmap bmp( wxGetApp().m_il_logo->GetSize() ) ;
   /*----------------------------------------------------------------------*/
   wxMemoryDC dc( bmp ) ;
   dc.SetBackground( color_background ) ;
   dc.Clear() ;
   dc.SetFont( wxSystemSettings::GetFont( wxSYS_DEFAULT_GUI_FONT ) ) ;
   dc.SetTextForeground( color_foreground ) ;
   /*----------------------------------------------------------------------*/
   mtf::draw_wrapped( dc, s_name ) ;
   dc.SelectObject( wxNullBitmap ) ;
   /*----------------------------------------------------------------------*/
   return bmp ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxBitmap CChannel::load_logo()
{
   /*----------------------------------------------------------------------*/
   if( m_s_logo_fn.empty() ) { return wxBitmap() ; }
   /*----------------------------------------------------------------------*/
   wxLogNull no_log ; // Suppress useless warnings
   wxBitmap  bmp( m_s_logo_fn, wxBITMAP_TYPE_PNG ) ;
   /*--( Invalid file or invalid image size )------------------------------*/
   if( !bmp.IsOk() || bmp.GetSize() != wxGetApp().m_il_logo->GetSize() )
   {  bmp = build_logo() ; }
   else
   if( bmp.HasAlpha() )
   {
      /*--------------------------------------------------------------------+
      ! Remove alpha channel to improve logo display "quality" on different !
      ! backgrounds : channel list and media message                        !
      ! The playlist active record looks nicer                              !
      +--------------------------------------------------------------------*/
      wxBitmap bmp_no_alpha( bmp.GetSize() ) ;
      /*-------------------------------------------------------------------*/
      wxMemoryDC dc( bmp_no_alpha ) ;
      dc.SetBackground( wxSystemSettings::GetColour( wxSYS_COLOUR_WINDOW ) );
      dc.Clear() ;
      dc.DrawBitmap( bmp, 0, 0 ) ;
      dc.SelectObject( wxNullBitmap ) ;
      /*-------------------------------------------------------------------*/
      bmp = bmp_no_alpha ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return bmp ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CChannel::init_logo()
{
   /*--( If the file is unknown the logo is "built" )----------------------*/
   wxBitmap bmp = ( m_s_logo_fn.empty() ? build_logo() : load_logo() ) ;
   /*--( Add overlay for freebox "en option" channels )--------------------*/
   if( is_fbtvo() ) { draw_en_option_overlay( bmp ) ; }
   /*--( Replace or add logo ? )-------------------------------------------*/
   if( m_i_il_pos < 0 )
   {  m_i_il_pos = wxGetApp().m_il_logo->Add( bmp ) ; }
   else
   {  wxGetApp().m_il_logo->Replace( m_i_il_pos, bmp ) ; }
   /*----------------------------------------------------------------------*/
   m_boo_logo_ok = true ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannel::get_logo_il_pos()
{
   /*----------------------------------------------------------------------*/
   if( !m_boo_logo_ok ) { init_logo() ; }
   /*----------------------------------------------------------------------*/
   return m_i_il_pos ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxBitmap CChannel::get_logo()
{
   /*----------------------------------------------------------------------*/
   return wxGetApp().m_il_logo->GetBitmap( get_logo_il_pos() ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannel::find_stream( const wxString &s_name ) const
{
   /*----------------------------------------------------------------------*/
   if( s_name.empty() ) { return -1 ; }
   /*----------------------------------------------------------------------*/
   auto it = std::find( m_vec_stream.begin(), m_vec_stream.end(), s_name ) ;
   return it == m_vec_stream.end()
          ? -1 : static_cast< int >( it - m_vec_stream.begin() ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
const str_stream *CChannel::get_stream()
{
   /*----------------------------------------------------------------------*/
   if( is_refresh_streams_needed() ) { return nullptr ; }
   /*--( By default the active stream is the first one )-------------------*/
   int  i_stream = find_stream( m_s_stream_name ) ;
   bool boo_stream_reinit = ( i_stream < 0 ) ;
   if( boo_stream_reinit ) { i_stream = 0 ; }
   /*----------------------------------------------------------------------*/
   str_stream *p_stream = &m_vec_stream[ i_stream ] ;
   /*--( If any, the change has to be kept )-------------------------------*/
   if( boo_stream_reinit )
   {  m_s_stream_name = p_stream->m_s_name ;
      set_save_cfg( true ) ;
   }
   /*----------------------------------------------------------------------*/
   return p_stream ;
   /*----------------------------------------------------------------------*/
}

/*--( "Fixed" for not "dynamic stream" like the ones from live channels )--*/
void CChannel::init_fixed_streams()
{
   /*----------------------------------------------------------------------*/
   if( m_vec_stream.empty() ) { return ; }
   /*--( Streams sort and default : we are only sure of the freebox ones )-*/
   if( m_pl_known == PK_FBTV )
   {
      /*-------------------------------------------------------------------*/
      std::sort( m_vec_stream.begin(), m_vec_stream.end(),
                 []( const str_stream &s1, const str_stream &s2 )
                 {  /*-----------------------------------------------------*/
                    int i_q1 = str_stream::st_get_quality_ind( s1.m_s_name );
                    int i_q2 = str_stream::st_get_quality_ind( s2.m_s_name );
                    /*-----------------------------------------------------*/
                    if( i_q1 >= 0 && i_q2 >= 0 ) { return i_q1 < i_q2 ; }
                    /*-----------------------------------------------------*/
                    if( i_q1 >= 0 ) { return true ; }
                    /*-----------------------------------------------------*/
                    if( i_q2 >= 0 ) { return false ; }
                    /*--( None of the names are quality => sort on names )-*/
                    return s1.m_s_name < s2.m_s_name ;
                    /*-----------------------------------------------------*/
                 }
               ) ;
      /*-------------------------------------------------------------------*/
   }

   /*--( The default quality is searched based on "Preferences" )----------*/
   wxString s_quality = ( wxGetApp().M_s_def_quality.empty()
                          ? "auto" : wxGetApp().M_s_def_quality.get()
                        ) ;
     t_vec_stream_it it_stream
   = std::find_if( m_vec_stream.begin(), m_vec_stream.end(),
                   [ &s_quality ]( const str_stream &stream )
                   {  return stream.m_s_name == s_quality ; }
                 ) ;
   /*--( If the default quality is not present, the first stream is used )-*/
   if( it_stream == m_vec_stream.end() )
   {  it_stream = m_vec_stream.begin() ; }
   m_s_stream_name = it_stream->m_s_name ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CChannel::datas_to_string()
{
   /*----------------------------------------------------------------------*/
   return wxString::Format( "\"%s\" %d %d %d", m_s_stream_name,
                            m_i_video_track, m_i_audio_track, m_i_video_spu
                          ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CChannel::string_to_datas( const wxString &s_val )
{
   /*----------------------------------------------------------------------*/
   char a_c_stream_name[ 256 ] ;
   int  i_video_track = DEFAULT_TRACK ;
   int  i_audio_track = DEFAULT_TRACK ;
   int  i_video_spu   = DEFAULT_TRACK ;
   /*----------------------------------------------------------------------*/
   wxSscanf( s_val, "\"%255[^\"]\" %d %d %d", a_c_stream_name,
             &i_video_track, &i_audio_track, &i_video_spu
           ) ;
   a_c_stream_name[ WXSIZEOF( a_c_stream_name ) - 1 ] = '\0' ;

   /*----------------------------------------------------------------------*/
   m_s_stream_name = a_c_stream_name ;
   m_i_video_track = i_video_track ;
   m_i_audio_track = i_audio_track ;
   m_i_video_spu   = i_video_spu   ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CChannel::rename( const wxString &s_new_name )
{
   /*----------------------------------------------------------------------*/
   set_short_name( s_new_name ) ;
   init_logo() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CChannel::get_vlc_init() const
{  return wxGetApp().M_s_vlc_init_other.get() ; }

/*-------------------------------------------------------------------------*/
int CChannel::add_stream( str_stream &stream )
{
   /*--( The stream name may not be specified ... )------------------------*/
   if( stream.m_s_name.empty() )
   {  stream.m_s_name.Printf( _( "stream %d" ),
                              static_cast< int >( m_vec_stream.size() ) + 1
                            ) ;
   }
   /*--( Append avoiding duplicates )--------------------------------------*/
   auto it = std::find_if( m_vec_stream.begin(), m_vec_stream.end(),
                           [ &stream ]( const str_stream &stream_curr )
                           {  return    stream_curr.m_s_url
                                     == stream.m_s_url ;
                           }
                         ) ;
   /*----------------------------------------------------------------------*/
   if( it != m_vec_stream.end() ) { return -1 ; }
   /*----------------------------------------------------------------------*/
   m_vec_stream.emplace_back( stream ) ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CChannel::reset_nb_viewers()
{
   /*----------------------------------------------------------------------*/
   csec_channel_enter() ;
   /*----------------------------------------------------------------------*/
   m_ti_nb_viewers = 0 ;
   m_l_nb_viewers  = 0 ;
   /*----------------------------------------------------------------------*/
   csec_channel_leave() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
long CChannel::get_nb_viewers()
{
   /*----------------------------------------------------------------------*/
   csec_channel_enter() ;
   /*----------------------------------------------------------------------*/
   long l_nb_viewers = m_l_nb_viewers ;
   /*----------------------------------------------------------------------*/
   csec_channel_leave() ;
   /*----------------------------------------------------------------------*/
   return l_nb_viewers ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
bool CChannel::is_nb_viewers_refresh_needed()
{
   /*----------------------------------------------------------------------*/
   csec_channel_enter() ;
   /*----------------------------------------------------------------------*/
   bool boo_refresh_needed
      = (    wxGetApp().M_i_viewers_delay.get() > 0
          &&    time( nullptr ) - m_ti_nb_viewers
             >= wxGetApp().M_i_viewers_delay.get()
        ) ;
   /*----------------------------------------------------------------------*/
   csec_channel_leave() ;
   /*----------------------------------------------------------------------*/
   return boo_refresh_needed ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CChannel::message_refresh_streams( CVlcMedia *p_media )
{
   /*----------------------------------------------------------------------*/
   if( p_media == nullptr ) { return ; }
   /*--( GUI changes have to be done in the main thread )------------------*/
   wxQueueEvent( p_media,
                 new sc::CStringEvent( EVT_SET_MESSAGE,
                                       _( "Looking for streams ..." )
                                     )
               ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int st_download_main_url( const wxString &co_s_url,
                          wxString &s_data, wxString *p_s_err_msg
                        )
{
   /*----------------------------------------------------------------------*/
   switch( sc::download( co_s_url, s_data ) )
   {
      /*-------------------------------------------------------------------*/
      case sc::CDownloadBase::DE_NO_ERROR_DONE :
         break ;
      /*-------------------------------------------------------------------*/
      case sc::CDownloadBase::DE_HOST_NOT_FOUND :
         /*----------------------------------------------------------------*/
         if( p_s_err_msg != nullptr )
         {  *p_s_err_msg = _( "Invalid URL" ) ; }
         /*----------------------------------------------------------------*/
         return -1 ;
      /*-------------------------------------------------------------------*/
      case sc::CDownloadBase::DE_FILE_NOT_FOUND :
         /*----------------------------------------------------------------*/
         if( p_s_err_msg != nullptr )
         {    *p_s_err_msg
            = CChannelLive::st_get_gs_error_message(
                                          CChannelLive::GSERR_UNKNOWN_CHANNEL
                                                   ) ;
         }
         /*----------------------------------------------------------------*/
         return -2 ;
      /*-------------------------------------------------------------------*/
      default :
         /*----------------------------------------------------------------*/
         if( p_s_err_msg != nullptr )
         {  *p_s_err_msg = _( "Problem downloading data" ) ; }
         /*----------------------------------------------------------------*/
         return -3 ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
static int st_extract_channel_id_from_url( wxString &s_url,
                                           wxString &s_id,
                                           wxString *p_s_err_msg
                                         )
{
   /*--( The URL may have to be un-shortened to extract the channel id )---*/
   if( mtf::unshort_url( s_url, p_s_err_msg ) != 0 ) { return -1 ; }

   /*--( The channel id is the last element of the url )-------------------*/
   s_id = s_url.AfterLast( '/' ).BeforeFirst( '?' ) ;
   if( s_id.empty() )
   {  /*-------------------------------------------------------------------*/
      if( p_s_err_msg != nullptr )
      {  *p_s_err_msg = _( "The URL should end with the channel id" ) ; }
      /*-------------------------------------------------------------------*/
      return -2 ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
static int st_extract_name_from_page( const wxString &s_data,
                                      wxString &s_name,
                                      wxString *p_s_err_msg
                                    )
{  /*--( Manage different arrangements of "meta property content" )--------*/
   wxString s_meta ;
   if(    sc::match( s_data,
                     R"=(<meta\M([^>]*?\mproperty=["']og:title["'].*?)>)=",
                     &s_meta
                   ) <= 1
       || sc::match( s_meta, R"=(\mcontent=["']([^"']*?)["'])=",
                     &s_name
                   ) <= 1
     )
   {  /*-------------------------------------------------------------------*/
      if( p_s_err_msg != nullptr )
      {  *p_s_err_msg = _( "The channel name could not be found" ) ; }
      /*-------------------------------------------------------------------*/
      return -1 ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   s_name = mtf::unhtml( s_name ).Trim().Trim( false ) ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CChannelDynamic::st_get_gs_error_message( e_gs_error gse )
{
   /*----------------------------------------------------------------------*/
   switch( gse )
   {
      /*-------------------------------------------------------------------*/
      case GSERR_NO_ERROR : return wxEmptyString ;
      case GSERR_DOWNLOAD : return _( "Problem downloading data" ) ;
      case GSERR_UNKNOWN_CHANNEL : return _( "Unknown channel" ) ;
      case GSERR_PARSING : return _( "Information parsing error" ) ;
      case GSERR_RETRY_LATER : return wxEmptyString ;
      case GSERR_FORBIDDEN : return _( "Access refused" ) ;
      case GSERR_PASSWORD_PROTECTED : return _( "Password protected" ) ;
      case GSERR_NOT_VIEWABLE_HERE :
         return wxString::Format( _( "The owner of this channel doesn't "
                                     "want it to be watched using an "
                                     "application like %s"
                                   ),
                                  APP_DISPLAY_NAME
                                ) ;
      /*-------------------------------------------------------------------*/
      case GSERR_UNHANDLED : return _( "Unhandled error" ) ;
      /*-------------------------------------------------------------------*/
      default :
         return wxString::Format( _( "Unknown error code (%s)" ), gse ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
bool CChannelDynamic::is_refresh_streams_needed() const
{
   /*--( No streams or streams older than xx minutes )---------------------*/
   return m_vec_stream.empty() || time( nullptr ) - m_ti_last_gs >= 10 * 60 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CChannelDynamic::reset_streams()
{
   /*----------------------------------------------------------------------*/
   wxCriticalSectionLocker enter_csec( m_csec_channel ) ;
   m_vec_stream.clear() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelDynamic::hls_extract_streams( const wxString &s_url_m3u )
{
   /*----------------------------------------------------------------------*/
   m_vec_stream.clear() ;
   /*--( Download the m3u )------------------------------------------------*/
   wxString s_data ;
   wxString s_last_url ;
   int i_err = sc::download( s_url_m3u, s_data, &s_last_url ) ;
   if( i_err == sc::CDownloadBase::DE_FORBIDDEN )
   {  set_gs_error( GSERR_FORBIDDEN ) ; return -1 ; }
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  set_gs_error( has_retry() ? GSERR_RETRY_LATER : GSERR_DOWNLOAD ) ;
      return -2 ;
   }

   /*----------------------------------------------------------------------*/
   wxStringInputStream is( s_data ) ;
   wxTextInputStream   tis( is ) ;
   wxString            s_line ;
   int                 i_line_num = 0 ;
   wxString   s_url_video = s_last_url.BeforeFirst( '?' ).BeforeLast( '/' ) ;
   /*----------------------------------------------------------------------*/
   while( !tis.GetInputStream().Eof() )
   {
      /*-------------------------------------------------------------------*/
      wxString s_name ;
      wxString s_url ;
      /*-------------------------------------------------------------------*/
      while(    !tis.GetInputStream().Eof()
             && ( s_line = mtf::read_m3u_line( tis, i_line_num )
                ).StartsWith( '#' )
           )
      {  /*----------------------------------------------------------------*/
         if( s_line.StartsWith( "#EXT-X-STREAM-INF:" ) )
         {  sc::match( s_line, R"=(\mRESOLUTION=(.*?)\M)=", &s_name ) ; }
         else /*--( Used by Twitch )---------------------------------------*/
         if( s_line.StartsWith( "#EXT-X-MEDIA:" ) )
         {  sc::match( s_line, R"=(\mNAME="(.*?)")=", &s_name ) ; }
         /*--( No name but bandwidth )-------------------------------------*/
         if( s_name.empty() )
         {
            /*-------------------------------------------------------------*/
            wxString s_bw ;
            /*--( If set, try to express bandwidth in in a bigger unit )---*/
            if(    sc::match( s_line, R"=(\mBANDWIDTH=(\d*))=", &s_bw ) > 1
                && s_bw.EndsWith( "000" )
              )
            {  s_bw = s_bw.Left( s_bw.size() - 3 ) + 'k' ; }
            /*--( Audio "only" should not have a frame rate info (RT fr) )-*/
            if( sc::match( s_line, R"=(\mFRAME-RATE=)=" ) == 0 )
            {  s_name = s_bw ; }
            else
            {  s_name = "audio" ;
               if( !s_bw.empty() ) { s_name += ' ' + s_bw ; }
            }
            /*-------------------------------------------------------------*/
         }
         /*----------------------------------------------------------------*/
      }
      /*--( We should have found the URL )---------------------------------*/
      if( !tis.GetInputStream().Eof() )
      {  /*--( The URL may be relative )-----------------------------------*/
         if( !s_line.StartsWith( "http" ) ) { s_url = s_url_video + '/' ; }
         /*----------------------------------------------------------------*/
         s_url += s_line ;
         /*--( Each stream is accepted only once )-------------------------*/
         if(    !s_name.empty()
             && !s_url.empty()
             && find_stream( s_name ) < 0
           )
         {  m_vec_stream.emplace_back( s_name, s_url ) ; }
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   }

   /*--( Normally we should have found something )-------------------------*/
   if( m_vec_stream.empty() )
   {  set_gs_error( GSERR_PARSING ) ; return -3 ; }

   /*-----------------------------------------------------------------------+
   ! Sort streams in descending quality order. The names are supposed to be !
   ! formatted "width x height". Considering that all streams have the same !
   ! aspect ratio, sorting on only the first number is ok                   !
   +-----------------------------------------------------------------------*/
   std::sort( m_vec_stream.begin(), m_vec_stream.end(),
              []( const str_stream &s1, const str_stream &s2 )
              {  return wxAtoi( s1.m_s_name ) > wxAtoi( s2.m_s_name ) ; }
            ) ;
   /*-----------------------------------------------------------------------+
   ! Add hls stream itself. In previous versions I didn't include it        !
   ! because I felt it was'nt a very good user experience. Now it seems ok. !
   ! I add it but at the end of the stream list.                            !
   +-----------------------------------------------------------------------*/
   add_hls_stream( s_url_m3u ) ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CChannelDynamic::add_hls_stream( const wxString &s_url_m3u )
{
   /*----------------------------------------------------------------------*/
   m_vec_stream.emplace_back( "HLS", s_url_m3u ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelDynamic::refresh_streams( CVlcMedia *p_media )
{
   /*----------------------------------------------------------------------*/
   wxCriticalSectionLocker enter_csec( m_csec_channel ) ;
   /*--( Useful to go further ? )------------------------------------------*/
   if( !is_refresh_streams_needed() ) { return 0 ; }
   /*----------------------------------------------------------------------*/
   set_gs_error( GSERR_NO_ERROR ) ;
   message_refresh_streams( p_media ) ;
   m_vec_stream.clear() ;
   time( &m_ti_last_gs ) ;
   /*----------------------------------------------------------------------*/
   return lookup_streams() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CChannelDailymotion::get_vlc_init() const
{  return wxGetApp().M_s_vlc_init_dm.get() ; }

/*-------------------------------------------------------------------------*/
int CChannelDailymotion::st_extract_data_from_url( const wxString &co_s_url,
                                                   wxString &s_id,
                                                   wxString &s_name,
                                                   wxString *p_s_err_msg
                                                 )
{
   /*--( The URL may have to be un-shortened to extract the channel id )---*/
   wxString s_url = co_s_url ;
   if( st_extract_channel_id_from_url( s_url, s_id, p_s_err_msg ) != 0 )
   {  return -1 ; }
   /*----------------------------------------------------------------------*/
   wxString s_data ;
   if( st_download_main_url( s_url, s_data, p_s_err_msg ) != 0 )
   {  return -2 ; }

   /*--( One way to check if the channel exists ... )----------------------*/
   if( s_data.find( "<meta property=\"og:url\"" ) == wxString::npos )
   {  /*-------------------------------------------------------------------*/
      if( p_s_err_msg != nullptr )
      {  *p_s_err_msg = st_get_gs_error_message( GSERR_UNKNOWN_CHANNEL ) ; }
      /*-------------------------------------------------------------------*/
      return -3 ;
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   if( st_extract_name_from_page( s_data, s_name, p_s_err_msg ) != 0 )
   {  return -4 ; }
   /*--( Clean name )------------------------------------------------------*/
   s_name.EndsWith( wxString::FromUTF8( u8"- Vid\u00e9o dailymotion" ),
                    &s_name
                  ) ;
   s_name = s_name.Trim() ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelDailymotion::lookup_streams()
{
   /*----------------------------------------------------------------------*/
   const char *co_p_c_url_fmt
            = "https://www.dailymotion.com/player/metadata/video/%s" ;
   const wxString co_s_url = wxString::Format( co_p_c_url_fmt, m_s_id ) ;
   wxString s_data ;
   /*----------------------------------------------------------------------*/
   int i_err = sc::download( co_s_url, s_data ) ;
   if( i_err == sc::CDownloadBase::DE_FILE_NOT_FOUND )
   {  set_gs_error( GSERR_UNKNOWN_CHANNEL ) ; return -1 ; }
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  set_gs_error( GSERR_DOWNLOAD ) ; return -2 ; }
   /*--( Forbidden access ? )----------------------------------------------*/
   if(    s_data.Find( R"=("code":"403",)=" ) != wxNOT_FOUND
       && s_data.Find( R"=("type":"password_protected","error_data":{)="
                       R"=("reason":"video_password_protected")="
                     ) != wxNOT_FOUND
     )
   {  set_gs_error( GSERR_PASSWORD_PROTECTED ) ; return -3 ; }
   /*--( Looking for the "auto" stream )-----------------------------------*/
   wxString s_url ;
   if( sc::match( s_data,
                  R"=("qualities":{"auto":\[{)="
                  R"=("type":"application\\/x-mpegURL","url":"(.*?)")=",
                  &s_url
                ) <= 1
     )
   {  set_gs_error( GSERR_RETRY_LATER ) ; return -4 ; }
   /*----------------------------------------------------------------------*/
   s_url.Replace( R"=(\/)=", "/", true ) ;
   /*----------------------------------------------------------------------*/
   if( hls_extract_streams( s_url ) != 0 ) { return -5 ; }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelDailymotion::refresh_nb_viewers()
{
   /*----------------------------------------------------------------------*/
   wxCriticalSectionLocker enter_csec( m_csec_channel ) ;
   /*----------------------------------------------------------------------*/
   time( &m_ti_nb_viewers ) ;
   /*----------------------------------------------------------------------*/
   const char *co_p_c_url_fmt
                   = "https://api.dailymotion.com/video/%s?fields=audience" ;
   const wxString co_s_url = wxString::Format( co_p_c_url_fmt, m_s_id ) ;
   /*----------------------------------------------------------------------*/
   wxString s_data ;
   int i_err = sc::download( co_s_url, s_data ) ;
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  return -1 ; }
   /*----------------------------------------------------------------------*/
   wxString s_nb ;
   if(    sc::match( s_data, R"=("audience":(\d*))=", &s_nb ) <= 1
       ||  s_nb.empty()
       || !s_nb.ToLong( &m_l_nb_viewers )
     )
   {  return -2 ; }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelDailymotion::refresh_onair()
{
   /*----------------------------------------------------------------------*/
   wxCriticalSectionLocker enter_csec( m_csec_channel ) ;
   /*----------------------------------------------------------------------*/
   m_boo_onair = false ;
   /*----------------------------------------------------------------------*/
   const char *co_p_c_url_fmt
                   = "https://api.dailymotion.com/video/%s?fields=onair" ;
   const wxString co_s_url = wxString::Format( co_p_c_url_fmt, m_s_id ) ;
   /*----------------------------------------------------------------------*/
   wxString s_data ;
   int i_err = sc::download( co_s_url, s_data ) ;
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  return -1 ; }
   /*----------------------------------------------------------------------*/
   wxString s_onair ;
   if(    sc::match( s_data, R"=("onair":(\w*))=", &s_onair ) <= 1
       || s_onair.empty()
     )
   {  return -2 ; }
   /*----------------------------------------------------------------------*/
   m_boo_onair = ( s_onair == "true" ) ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
sc::t_lst_string CChannelTwitch::st_get_header()
{  return wxGetApp().m_co_ls_twitch_header ; }

/*-------------------------------------------------------------------------*/
wxString CChannelTwitch::get_vlc_init() const
{  return wxGetApp().M_s_vlc_init_tw.get() ; }

/*-------------------------------------------------------------------------*/
int CChannelTwitch::st_extract_data_from_url( const wxString &co_s_url,
                                              wxString &s_id,
                                              wxString &s_name,
                                              wxString *p_s_err_msg
                                            )
{
   /*--( The URL may have to be un-shortened to extract the channel id )---*/
   wxString s_url = co_s_url ;
   if( st_extract_channel_id_from_url( s_url, s_id, p_s_err_msg ) != 0 )
   {  return -1 ; }
   /*--( Some twitch services accept the id only lowercase )---------------*/
   s_id.MakeLower() ;

   /*----------------------------------------------------------------------*/
   s_url = "https://api.twitch.tv/helix/users?login=" + s_id ;
   /*----------------------------------------------------------------------*/
   wxString s_data ;
   int i_err = sc::download( s_url, s_data, nullptr, -1, st_get_header() ) ;
   if( i_err == sc::CDownloadBase::DE_FILE_NOT_FOUND )
   {  /*-------------------------------------------------------------------*/
      if( p_s_err_msg != nullptr )
      {  *p_s_err_msg = st_get_gs_error_message( GSERR_UNKNOWN_CHANNEL ) ; }
      /*-------------------------------------------------------------------*/
      return -2 ;
      /*-------------------------------------------------------------------*/
   }
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE ) { return -3 ; }

   /*--( Extract name )----------------------------------------------------*/
   if( sc::match( s_data, R"=("display_name":"(.*?)")=", &s_name ) <= 1 )
   {  /*-------------------------------------------------------------------*/
      if( p_s_err_msg != nullptr )
      {  *p_s_err_msg = st_get_gs_error_message( GSERR_UNKNOWN_CHANNEL ) ; }
      /*-------------------------------------------------------------------*/
      return -4 ;
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CChannelLive::e_gs_error
              CChannelTwitch::st_get_token_sig( const wxString &s_id,
                                                wxString *p_s_token,
                                                wxString *p_s_sig
                                              )
{
   /*----------------------------------------------------------------------*/
   const char *co_p_c_url_fmt
            = "https://api.twitch.tv/api/channels/%s/access_token?"
              "oauth_token=undefined&need_https=true&platform=web&"
              "player_type=site&player_backend=mediaplayer" ;
   const wxString co_s_url = wxString::Format( co_p_c_url_fmt, s_id ) ;
   /*----------------------------------------------------------------------*/
   wxString s_data ;
   int i_err = sc::download( co_s_url, s_data, nullptr, -1, st_get_header());
   /*----------------------------------------------------------------------*/
   if( i_err == sc::CDownloadBase::DE_FILE_NOT_FOUND )
   {  return GSERR_UNKNOWN_CHANNEL ; }
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  return GSERR_DOWNLOAD ; }

   /*--( Extract token and sig )-------------------------------------------*/
   if( sc::match( s_data, R"=("token":"(.*?)","sig":"(.*?)")=",
                  p_s_token, p_s_sig
                ) <= 2
     )
   {  return GSERR_PARSING ; }
   /*--( Reworking the token content )-------------------------------------*/
   if( p_s_token != nullptr )
   {  /*--------------------------------------------------------------------+
      ! The token must be carefully modified ... ! if not, the expected     !
      ! data download will be refused                                       !
      +--------------------------------------------------------------------*/
      wxRegEx( R"=(\\(.))=" ).Replace( p_s_token, R"=(\1)=" ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return GSERR_NO_ERROR ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelTwitch::lookup_streams()
{
   /*----------------------------------------------------------------------*/
   wxString s_token ;
   wxString s_sig ;

   /*--( Get the "base" information before extracting the streams )--------*/
   set_gs_error( st_get_token_sig( m_s_id, &s_token, &s_sig ) ) ;
   if( get_gs_error() != GSERR_NO_ERROR ) { return -1 ; }

   /*--( Streams )---------------------------------------------------------*/
   const char *co_p_c_url_fmt
       = "https://usher.ttvnw.net/api/channel/hls/%s.m3u8?"
         "player=twitchweb&p=%d&type=any"
         "&allow_source=true&allow_audio_only=true&token=%s&sig=%s" ;
   const wxString co_s_url
       = wxString::Format( co_p_c_url_fmt, m_s_id, mtf::rand( 0, 9999999 ),
                           mtf::url_encode( s_token ), s_sig
                         ) ;
   /*----------------------------------------------------------------------*/
   if( hls_extract_streams( co_s_url ) != 0 ) { return -2 ; }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelTwitch::refresh_nb_viewers()
{
   /*----------------------------------------------------------------------*/
   wxCriticalSectionLocker enter_csec( m_csec_channel ) ;
   /*----------------------------------------------------------------------*/
   time( &m_ti_nb_viewers ) ;
   /*----------------------------------------------------------------------*/
   const wxString co_s_url
               = "https://api.twitch.tv/helix/streams?user_login=" + m_s_id ;
   /*----------------------------------------------------------------------*/
   wxString s_data ;
   int i_err = sc::download( co_s_url, s_data, nullptr, -1, st_get_header());
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE ) { return -1 ; }
   /*----------------------------------------------------------------------*/
   wxString s_nb ;
   if(    sc::match( s_data, R"=("viewer_count":(\d*))=", &s_nb ) <= 1
       ||  s_nb.empty()
       || !s_nb.ToLong( &m_l_nb_viewers )
     )
   {  return -2 ; }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelTwitch::refresh_onair()
{
   /*----------------------------------------------------------------------*/
   wxCriticalSectionLocker enter_csec( m_csec_channel ) ;
   /*----------------------------------------------------------------------*/
   m_boo_onair = false ;
   /*----------------------------------------------------------------------*/
   const wxString co_s_url
               = "https://api.twitch.tv/helix/streams?user_login=" + m_s_id ;
   /*----------------------------------------------------------------------*/
   wxString s_data ;
   int i_err = sc::download( co_s_url, s_data, nullptr, -1, st_get_header());
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE ) { return -1 ; }
   /*----------------------------------------------------------------------*/
   m_boo_onair = !s_data.StartsWith( R"=({"data":[],)=" ) ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CChannelYoutube::get_vlc_init() const
{  return wxGetApp().M_s_vlc_init_yt.get() ; }

/*-------------------------------------------------------------------------*/
int CChannelYoutube::st_extract_data_from_url( const wxString &co_s_url,
                                               wxString &s_id,
                                               wxString &s_name,
                                               wxString *p_s_err_msg
                                             )
{
   /*----------------------------------------------------------------------*/
   wxString s_data ;
   if( st_download_main_url( co_s_url, s_data, p_s_err_msg ) != 0 )
   {  return -1 ; }

   /*--( Extract channel id )----------------------------------------------*/
   if(    sc::match( s_data,
                     R"=(<meta itemprop="channelId" content="(.*?)">)=",
                     &s_id
                   ) <= 1
       || s_id.empty()
     )
   {  /*-------------------------------------------------------------------*/
      if( p_s_err_msg != nullptr )
      {  *p_s_err_msg = st_get_gs_error_message( GSERR_UNKNOWN_CHANNEL ) ; }
      /*-------------------------------------------------------------------*/
      return -2 ;
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   if( st_extract_name_from_page( s_data, s_name, p_s_err_msg ) != 0 )
   {  return -3 ; }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelYoutube::lookup_streams()
{
   /*----------------------------------------------------------------------*/
   if( extract_live_video_id() != 0 )
   {  /*--( If it exists, the channel is not currently streaming )---------*/
      if( get_gs_error() == GSERR_NO_ERROR )
      {  set_gs_error( GSERR_RETRY_LATER ) ; }
      /*-------------------------------------------------------------------*/
      return -2 ;
      /*-------------------------------------------------------------------*/
   }

   /*--( Get video live video information )--------------------------------*/
   wxString s_url
          = "https://youtube.com/get_video_info?video_id=" + m_s_video_id ;
   wxString s_data ;
   int i_err = sc::download( s_url, s_data ) ;
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  set_gs_error( GSERR_DOWNLOAD ) ; return -3 ; }

   /*--( Check youtube error with status ok ... (for example Bloomberg) )--*/
   if( sc::match( s_data, R"=(%22status%22%3A%22UNPLAYABLE%22)=" ) == 1 )
   {  set_gs_error( GSERR_NOT_VIEWABLE_HERE ) ;
      return -4 ;
   }
   /*--( Check another youtube error code )--------------------------------*/
   if( sc::match( s_data, R"=(\mstatus=fail\M)=" ) == 1 )
   {  /*--( Extract error code )-------------------------------------------*/
      wxString s_err_code ;
      if( sc::match( s_data, R"=(\merrorcode=(\d+?)\M)=", &s_err_code) <= 1 )
      {  set_gs_error( GSERR_PARSING ) ; return -5 ; }
      /*--( Video can only be watched on Youtube ? )-----------------------*/
      if( s_err_code == "150" )
      {  set_gs_error( GSERR_NOT_VIEWABLE_HERE ) ; }
      else
      {  set_gs_error( GSERR_UNHANDLED ) ; }
      /*-------------------------------------------------------------------*/
      return -6 ;
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   s_data = wxURI::Unescape( s_data ) ;

   /*--( The live is only scheduled ? )------------------------------------*/
   if( s_data.Find( R"=("scheduledStartTime")=" ) != wxNOT_FOUND )
   {  set_gs_error( GSERR_RETRY_LATER ) ; return -7 ; }

   /*--( Looking for the m3u url : new and old way )-----------------------*/
   wxString s_url_hls ;
   if(    sc::match( s_data, R"=("hlsManifestUrl":"(.*?)")=", &s_url_hls
                   ) <= 1
       && sc::match( s_data, R"=(\&hlsvp=(.*?)\&)=", &s_url_hls
                   ) <= 1
     )
   {  set_gs_error( GSERR_PARSING ) ; return -8 ; }

   /*----------------------------------------------------------------------*/
   if( hls_extract_streams( s_url_hls ) != 0 ) { return -9 ; }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*--------------------------------------------------------------------------+
! Starting 2019/09 it seems the "live_stats" URL doesn't work anymore :     !
! https://www.youtube.com/live_stats?v=" + m_s_video_id ;                   !
! Therefore it has to be done "manually"                                    !
+--------------------------------------------------------------------------*/
int CChannelYoutube::refresh_nb_viewers()
{
   /*----------------------------------------------------------------------*/
   wxASSERT( !m_s_video_id.empty() ) ;
   /*----------------------------------------------------------------------*/
   wxCriticalSectionLocker enter_csec( m_csec_channel ) ;
   /*----------------------------------------------------------------------*/
   time( &m_ti_nb_viewers ) ;
   /*----------------------------------------------------------------------*/
   const wxString co_s_url
                = "https://www.youtube.com/watch?v=" + m_s_video_id ;
   /*----------------------------------------------------------------------*/
   wxString s_data ;
   int i_err = sc::download( co_s_url, s_data ) ;
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE ) { return -2 ; }

   /*--( Extract the viewer count message. It can be in any language )-----*/
   wxString s_text_viewcount ;
   if( sc::match( s_data,
            R"=(\{\\"viewCount\\":\{\\"runs\\":\[\{\\"text\\":\\"(.*?)\\")=",
                  &s_text_viewcount
                ) <= 1
     )
   {  return -4 ; }

   /*--( Extract the number from the text )--------------------------------*/
   wxString s_viewcount ;
   if( sc::match( s_text_viewcount, R"=(([0-9]+))=", &s_viewcount ) <= 1 )
   {  return -5 ; }

   /*----------------------------------------------------------------------*/
   if( !s_viewcount.ToLong( &m_l_nb_viewers ) ) { return -6 ; }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*--------------------------------------------------------------------------+
! Starting 2019/09 the youtube api "search" channeliId doesn't work         !
! anymore. The search is done  "manually"                                   !
+--------------------------------------------------------------------------*/
int CChannelYoutube::extract_live_video_id()
{
   /*----------------------------------------------------------------------*/
   wxCriticalSectionLocker enter_csec( m_csec_channel ) ;
   /*----------------------------------------------------------------------*/
   m_s_video_id.clear() ;
   /*----------------------------------------------------------------------*/
   const wxString co_s_url = get_url() ;
   /*----------------------------------------------------------------------*/
   wxString s_data ;
   int i_err = sc::download( co_s_url, s_data ) ;
   if( i_err == sc::CDownloadBase::DE_FILE_NOT_FOUND )
   {  set_gs_error( GSERR_UNKNOWN_CHANNEL ) ; return -1 ; }
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  set_gs_error( GSERR_DOWNLOAD ) ; return -2 ; }

   /*-----------------------------------------------------------------------+
   ! We only extract the first live stream : many channels can have more.   !
   ! REMEMBER when downloading no user agent is specified. This greatly     !
   ! lowers the weight of the downloaded file (at least for youtube)        !
   ! So ... don't base code on page source fetched using a browser.         !
   +------------------------------------------------------------------------+
   ! Solutions tested :                                                     !
   ! - Search "hqdefault_live.jpg" : used for upcoming live video too => NO !
   ! - The best seems to search "yt-badge-live" which comes after video id  !
   !    - Using only one regex is not easy because of the upcoming video    !
   !    - For now the simplest and effective seems to be non regex search   !
   +-----------------------------------------------------------------------*/

   /*--( Looking forward for a currently live video )----------------------*/
   const size_t sz_pos_badge_live = s_data.find( R"=( yt-badge-live)=" ) ;
   if( sz_pos_badge_live == wxString::npos ) { return -3 ; }

   /*--( Looking backward for the live video id )--------------------------*/
   const wxString co_s_watch = R"=( href="/watch?v=)=" ;
   const size_t sz_pos_watch = s_data.rfind( co_s_watch, sz_pos_badge_live );
   if( sz_pos_watch == wxString::npos ) { return -4 ; }

   /*--( Finally extracting the video id )---------------------------------*/
   const size_t sz_pos_start_id = sz_pos_watch + co_s_watch.size() ;
   const size_t sz_pos_end_id   = s_data.find( '"', sz_pos_start_id ) ;
   m_s_video_id = s_data.substr( sz_pos_start_id,
                                 sz_pos_end_id - sz_pos_start_id
                               ) ;
   /*----------------------------------------------------------------------*/
   if( m_s_video_id.empty() ) { return -5 ; }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelYoutube::refresh_onair()
{
   /*----------------------------------------------------------------------*/
   wxCriticalSectionLocker enter_csec( m_csec_channel ) ;
   /*----------------------------------------------------------------------*/
   m_boo_onair = false ;
   /*----------------------------------------------------------------------*/
   if( extract_live_video_id() != 0 ) { return -1 ; }
   /*----------------------------------------------------------------------*/
   m_boo_onair = true ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CChannelUstream::get_vlc_init() const
{  return wxGetApp().M_s_vlc_init_us.get() ; }

/*-------------------------------------------------------------------------*/
int CChannelUstream::st_extract_data_from_url( const wxString &co_s_url,
                                               wxString &s_id,
                                               wxString &s_name,
                                               wxString *p_s_err_msg
                                             )
{
   /*----------------------------------------------------------------------*/
   wxString s_data ;
   if( st_download_main_url( co_s_url, s_data, p_s_err_msg ) != 0 )
   {  return -1 ; }

   /*--( Extract channel id )----------------------------------------------*/
   if(    sc::match( s_data,
                   R"=(<meta name="ustream:channel_id" content="([^"]*)">)=",
                     &s_id
                   ) <= 1
       || s_id.empty()
     )
   {  /*-------------------------------------------------------------------*/
      if( p_s_err_msg != nullptr )
      {  *p_s_err_msg = st_get_gs_error_message( GSERR_UNKNOWN_CHANNEL ) ; }
      /*-------------------------------------------------------------------*/
      return -2 ;
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   if( st_extract_name_from_page( s_data, s_name, p_s_err_msg ) != 0 )
   {  return -3 ; }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelUstream::download_channel_json( wxString &s_data ) const
{
   /*----------------------------------------------------------------------*/
   const wxString s_url
                = "https://api.ustream.tv/channels/" + m_s_id + ".json" ;
   /*----------------------------------------------------------------------*/
   return sc::download( s_url, s_data ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CChannelUstream::get_url() const
{
   /*----------------------------------------------------------------------*/
   wxString s_data ;
   if(    download_channel_json( s_data )
       != sc::CDownloadBase::DE_NO_ERROR_DONE
     )
   {  return wxEmptyString ; }
   /*----------------------------------------------------------------------*/
   wxString s_url ;
   if( sc::match( s_data, R"=("tinyurl":"([^"]*)")=", &s_url ) <= 1 )
   {  return wxEmptyString ; }
   /*----------------------------------------------------------------------*/
   s_url.Replace( R"=(\/)=", "/", true ) ;
   /*----------------------------------------------------------------------*/
   return s_url ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelUstream::lookup_url_hls( wxString &s_url_hls )
{
   /*----------------------------------------------------------------------*/
   wxString s_data ;
   /*----------------------------------------------------------------------*/
   int i_err = download_channel_json( s_data ) ;
   if( i_err == sc::CDownloadBase::DE_FILE_NOT_FOUND )
   {  set_gs_error( GSERR_UNKNOWN_CHANNEL ) ; return -1 ; }
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  set_gs_error( GSERR_DOWNLOAD ) ; return -2 ; }

   /*--( Currently live ? )------------------------------------------------*/
   if( s_data.Find( R"=("status":"live",)=" ) == wxNOT_FOUND )
   {  set_gs_error( GSERR_RETRY_LATER ) ; return -4 ; }
   /*----------------------------------------------------------------------*/
   if( sc::match( s_data, R"=("hls":"([^"]*)")=", &s_url_hls ) <= 1 )
   {  set_gs_error( GSERR_PARSING ) ; return -5 ; }
   /*----------------------------------------------------------------------*/
   s_url_hls.Replace( R"=(\/)=", "/", true ) ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelUstream::lookup_streams()
{
   /*----------------------------------------------------------------------*/
   wxString s_url_hls ;
   /*----------------------------------------------------------------------*/
   if( lookup_url_hls( s_url_hls ) != 0 ) { return -1 ; }
   /*--( Ustream has separated audio/video streams )-----------------------*/
   add_hls_stream( s_url_hls ) ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelUstream::refresh_nb_viewers()
{
   /*----------------------------------------------------------------------*/
   wxCriticalSectionLocker enter_csec( m_csec_channel ) ;
   /*----------------------------------------------------------------------*/
   time( &m_ti_nb_viewers ) ;
   /*----------------------------------------------------------------------*/
   wxString s_data ;
   if(    download_channel_json( s_data )
       != sc::CDownloadBase::DE_NO_ERROR_DONE
     )
   {  return -1 ; }
   /*----------------------------------------------------------------------*/
   wxString s_nb ;
   if(    sc::match( s_data, R"=("viewer":(\d*))=", &s_nb ) <= 1
       ||  s_nb.empty()
       || !s_nb.ToLong( &m_l_nb_viewers )
     )
   {  return -2 ; }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelUstream::refresh_onair()
{
   /*----------------------------------------------------------------------*/
   wxCriticalSectionLocker enter_csec( m_csec_channel ) ;
   /*----------------------------------------------------------------------*/
   m_boo_onair = false ;
   /*----------------------------------------------------------------------*/
   wxString s_data ;
   if(    download_channel_json( s_data )
       != sc::CDownloadBase::DE_NO_ERROR_DONE
     )
   {  return -1 ; }
   /*----------------------------------------------------------------------*/
   m_boo_onair = ( s_data.Find( R"=("status":"live")=" ) != wxNOT_FOUND ) ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CChannelPeriscope::get_vlc_init() const
{  return wxGetApp().M_s_vlc_init_pe.get() ; }

/*-------------------------------------------------------------------------*/
int CChannelPeriscope::st_extract_data_from_url( const wxString &co_s_url,
                                                 wxString &s_id,
                                                 wxString &s_name,
                                                 wxString *p_s_err_msg
                                               )
{
   /*--( The URL may have to be un-shortened to extract the channel id )---*/
   wxString s_url = co_s_url ;
   if( mtf::unshort_url( s_url, p_s_err_msg ) != 0 ) { return -1 ; }

   /*--( The URL first "branch" should be the user name )------------------*/
   if(    sc::match( s_url, R"=(.*://.*?/(.*?)/?)=", &s_id ) <= 1
       || s_id.empty()
     )
   {  /*-------------------------------------------------------------------*/
      if( p_s_err_msg != nullptr )
      {  *p_s_err_msg = _( "The URL should contain the user name" ) ; }
      /*-------------------------------------------------------------------*/
      return -2 ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   wxString s_data ;
   int i_err = sc::download( s_url, s_data ) ;
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  /*-------------------------------------------------------------------*/
      if( p_s_err_msg != nullptr )
      {  *p_s_err_msg = st_get_gs_error_message( GSERR_DOWNLOAD ) ; }
      /*-------------------------------------------------------------------*/
      return -3 ;
      /*-------------------------------------------------------------------*/
   }

   /*--( The standard is not used by periscope )---------------------------*/
   if( sc::match( s_data, R"=(;display_name\&quot;:\&quot;(.*?)\&quot;)=",
                  &s_name
                ) <= 1
     )
   {  s_name = s_id ; }
   else
   {  s_name = mtf::unhtml( s_name ) ; }

   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelPeriscope::lookup_url_hls( wxString &s_url_hls )
{
   /*----------------------------------------------------------------------*/
   wxString s_url = get_url() ;
   wxString s_data ;
   int i_err = sc::download( s_url, s_data ) ;
   if( i_err == sc::CDownloadBase::DE_FILE_NOT_FOUND )
   {  set_gs_error( GSERR_UNKNOWN_CHANNEL ) ; return -1 ; }
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  set_gs_error( GSERR_DOWNLOAD ) ; return -2 ; }

   /*--( First extract ids )-----------------------------------------------*/
   if(    sc::match( s_data,
                     R"=(\&quot;public\&quot;:)="
                     R"=({\&quot;broadcastHistory\&quot;:{\&quot;token)="
                     R"=(\&quot;:{\&quot;session_id\&quot;:)="
                     R"=(\&quot;(.*?)\&quot;)="
                     R"=(.*?)="
                     R"=(\&quot;UserCache\&quot;:{\&quot;users\&quot;:)="
                     R"=({\&quot;.*?\&quot;:{\&quot;timestamp\&quot;:.*?,)="
                     R"=(\&quot;user\&quot;:{\&quot;class_name\&quot;:)="
                     R"=(\&quot;User\&quot;,\&quot;id\&quot;:)="
                     R"=(\&quot;(.*?)\&quot;)=",
                     &m_s_session_id, &m_s_user_id
                   ) <= 2
       || m_s_session_id.empty()
       || m_s_user_id.empty()
     )
   {  set_gs_error( GSERR_PARSING ) ; return -3 ; }

   /*--( Based on these ids, get the broadcast id )------------------------*/
   const char *co_p_c_url_fmt
            = "https://api.periscope.tv/api/v2/getUserBroadcastsPublic?"
              "user_id=%s&session_id=%s" ;
   s_url = wxString::Format( co_p_c_url_fmt, m_s_user_id, m_s_session_id ) ;
   i_err = sc::download( s_url, s_data ) ;
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  set_gs_error( GSERR_DOWNLOAD ) ; return -4 ; }
   /*----------------------------------------------------------------------*/
   wxString s_broadcast_id ;
   wxString s_broadcast_state ;
   if(    sc::match( s_data,
                     R"=(^{"broadcasts":\[{"class_name":)="
                     R"=("Broadcast","id":"(.*?)".*?"state":"(.*?)")=",
                     &s_broadcast_id, &s_broadcast_state
                   ) <= 2
       || s_broadcast_id.empty()
       || s_broadcast_state.empty()
     )
   {  set_gs_error( GSERR_PARSING ) ; return -5 ; }

   /*--( The stream is up ? )----------------------------------------------*/
   if( s_broadcast_state != "RUNNING" )
   {  set_gs_error( GSERR_RETRY_LATER ) ; return -6 ; }

   /*--( Loading the HLS )-------------------------------------------------*/
   s_url =   "https://api.periscope.tv/api/v2/getAccessPublic?broadcast_id="
           + s_broadcast_id ;
   /*----------------------------------------------------------------------*/
   i_err = sc::download( s_url, s_data ) ;
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  set_gs_error( GSERR_DOWNLOAD ) ; return -7 ; }
   /*----------------------------------------------------------------------*/
   if( sc::match( s_data, R"=("hls_url":"(.*?)")=", &s_url_hls ) <= 1 )
   {  set_gs_error( GSERR_RETRY_LATER ) ; return -8 ; }

   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelPeriscope::lookup_streams()
{
   /*----------------------------------------------------------------------*/
   wxString s_url_hls ;
   /*----------------------------------------------------------------------*/
   if( lookup_url_hls( s_url_hls ) != 0 ) { return -1 ; }
   /*----------------------------------------------------------------------*/
   add_hls_stream( s_url_hls ) ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelPeriscope::refresh_nb_viewers()
{
   /*----------------------------------------------------------------------*/
   if( m_s_user_id.empty() || m_s_session_id.empty() ) { return -1 ; }
   /*----------------------------------------------------------------------*/
   wxCriticalSectionLocker enter_csec( m_csec_channel ) ;
   /*----------------------------------------------------------------------*/
   time( &m_ti_nb_viewers ) ;
   /*--( The ids are not empty because we are here after play has started )*/
   const char *co_p_c_url_fmt
            = "https://api.periscope.tv/api/v2/getUserBroadcastsPublic?"
              "user_id=%s&session_id=%s" ;
   const wxString co_s_url
       = wxString::Format( co_p_c_url_fmt, m_s_user_id, m_s_session_id ) ;
   wxString s_data ;
   int i_err = sc::download( co_s_url, s_data ) ;
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE ) { return -2 ; }
   /*----------------------------------------------------------------------*/
   wxString s_nb ;
   if(    sc::match( s_data, R"=("n_total_watching":(\d*))=", &s_nb ) <= 1
       ||  s_nb.empty()
       || !s_nb.ToLong( &m_l_nb_viewers )
     )
   {  return -3 ; }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelPeriscope::refresh_onair()
{
   /*----------------------------------------------------------------------*/
   wxCriticalSectionLocker enter_csec( m_csec_channel ) ;
   wxString s_url_hls ;
   /*----------------------------------------------------------------------*/
   m_boo_onair = ( lookup_url_hls( s_url_hls ) == 0 ) ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CChannelMixer::get_vlc_init() const
{  return wxGetApp().M_s_vlc_init_mx.get() ; }

/*-------------------------------------------------------------------------*/
int CChannelMixer::st_extract_data_from_url( const wxString &co_s_url,
                                             wxString &s_id,
                                             wxString &s_name,
                                             wxString *p_s_err_msg
                                           )
{
   /*--( The URL may have to be un-shortened to extract the channel id )---*/
   wxString s_url = co_s_url ;
   if( st_extract_channel_id_from_url( s_url, s_id, p_s_err_msg ) != 0 )
   {  return -1 ; }

   /*----------------------------------------------------------------------*/
   const char *co_p_c_url_fmt
       = "https://mixer.com/api/v1/channels/%s?noCount=1" ;
   s_url.Printf( co_p_c_url_fmt, s_id ) ;
   wxString s_data ;
   int i_err = sc::download( s_url, s_data ) ;
   /*----------------------------------------------------------------------*/
   if( i_err == sc::CDownloadBase::DE_FILE_NOT_FOUND )
   {  /*-------------------------------------------------------------------*/
      if( p_s_err_msg != nullptr )
      {  *p_s_err_msg = st_get_gs_error_message( GSERR_UNKNOWN_CHANNEL ) ; }
      /*-------------------------------------------------------------------*/
      return -2 ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  /*-------------------------------------------------------------------*/
      if( p_s_err_msg != nullptr )
      {  *p_s_err_msg = st_get_gs_error_message( GSERR_DOWNLOAD ) ; }
      /*-------------------------------------------------------------------*/
      return -3 ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   s_name = s_id ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelMixer::download_channel_json( wxString &s_data ) const
{
   /*----------------------------------------------------------------------*/
   const char *co_p_c_url_fmt_id
       = "https://mixer.com/api/v1/channels/%s?noCount=1" ;
   wxString s_url = wxString::Format( co_p_c_url_fmt_id, m_s_id ) ;
   /*----------------------------------------------------------------------*/
   return sc::download( s_url, s_data ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelMixer::lookup_url_hls( wxString &s_url_hls )
{
   /*----------------------------------------------------------------------*/
   wxString s_data ;
   int i_err = download_channel_json( s_data ) ;
   if( i_err == sc::CDownloadBase::DE_FILE_NOT_FOUND )
   {  set_gs_error( GSERR_UNKNOWN_CHANNEL ) ; return -1 ; }
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  set_gs_error( GSERR_DOWNLOAD ) ; return -2 ; }

   /*--( Extract id )------------------------------------------------------*/
   wxString s_channel_id ;
   wxString s_online ;
   if(    sc::match( s_data, R"=("id":(.*?),.*?"online":(.*?),)=",
                     &s_channel_id, &s_online
                   ) <= 2
       || s_channel_id.empty()
       || s_online.empty()
     )
   {  set_gs_error( GSERR_PARSING ) ; return -3 ; }

   /*--( The stream is up ? )----------------------------------------------*/
   if( s_online != "true" )
   {  set_gs_error( GSERR_RETRY_LATER ) ; return -4 ; }

   /*--( Extract the hls url from this "manifest" )------------------------*/
   const char *co_p_c_url_fmt_hls
            = "https://mixer.com/api/v1/channels/%s/manifest.m3u8" ;
   s_url_hls = wxString::Format( co_p_c_url_fmt_hls, s_channel_id ) ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelMixer::lookup_streams()
{
   /*----------------------------------------------------------------------*/
   wxString s_url_hls ;
   /*----------------------------------------------------------------------*/
   if( lookup_url_hls( s_url_hls ) != 0 ) { return -1 ; }
   /*----------------------------------------------------------------------*/
   if( hls_extract_streams( s_url_hls ) != 0 ) { return -2 ; }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelMixer::refresh_nb_viewers()
{
   /*----------------------------------------------------------------------*/
   wxCriticalSectionLocker enter_csec( m_csec_channel ) ;
   /*----------------------------------------------------------------------*/
   time( &m_ti_nb_viewers ) ;
   /*----------------------------------------------------------------------*/
   wxString s_data ;
   if(    download_channel_json( s_data )
       != sc::CDownloadBase::DE_NO_ERROR_DONE
     )
   {  return -1 ; }
   /*----------------------------------------------------------------------*/
   wxString s_nb ;
   if(    sc::match( s_data, R"=("viewersCurrent":(\d*))=", &s_nb ) <= 1
       ||  s_nb.empty()
       || !s_nb.ToLong( &m_l_nb_viewers )
     )
   {  return -2 ; }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelMixer::refresh_onair()
{
   /*----------------------------------------------------------------------*/
   wxCriticalSectionLocker enter_csec( m_csec_channel ) ;
   wxString s_url_hls ;
   /*----------------------------------------------------------------------*/
   m_boo_onair = false ;
   /*----------------------------------------------------------------------*/
   wxString s_data ;
   if(    download_channel_json( s_data )
       != sc::CDownloadBase::DE_NO_ERROR_DONE
     )
   {  return -1 ; }
   /*----------------------------------------------------------------------*/
   m_boo_onair = ( s_data.Find( R"=("online":true,)=" ) != wxNOT_FOUND ) ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CChannelDynamicKnown::init_fnc_lookup_streams(
                                            CChannelDynamicKnown::e_cdtk cdtk
                                                  )
{
   /*----------------------------------------------------------------------*/
   switch( cdtk )
   {
      /*-------------------------------------------------------------------*/
      case CDTK_M3U :
           m_p_fnc_lookup_streams
         = &CChannelDynamicKnown::lookup_streams_m3u ;
         break ;
      /*-------------------------------------------------------------------*/
      case CDTK_FTV :
           m_p_fnc_lookup_streams
         = &CChannelDynamicKnown::lookup_streams_ftv ;
         break ;
      /*-------------------------------------------------------------------*/
      case CDTK_MYTF1 :
           m_p_fnc_lookup_streams
         = &CChannelDynamicKnown::lookup_streams_mytf1 ;
         break ;
      /*-------------------------------------------------------------------*/
      case CDTK_GULLI :
           m_p_fnc_lookup_streams
         = &CChannelDynamicKnown::lookup_streams_gulli ;
         break ;
      /*-------------------------------------------------------------------*/
      case CDTK_BFM :
           m_p_fnc_lookup_streams
         = &CChannelDynamicKnown::lookup_streams_bfm ;
         break ;
      /*-------------------------------------------------------------------*/
      case CDTK_MELODY :
           m_p_fnc_lookup_streams
         = &CChannelDynamicKnown::lookup_streams_melody ;
         break ;

      /*--( Avoid warning )------------------------------------------------*/
      default : break ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelDynamicKnown::lookup_streams_m3u()
{
   /*----------------------------------------------------------------------*/
   if( hls_extract_streams( m_s_url_m3u ) != 0 ) { return -1 ; }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*--------------------------------------------------------------------------+
! The code below is based on "catchuptvandmore"'s one                       !
+--------------------------------------------------------------------------*/
int CChannelDynamicKnown::lookup_streams_ftv()
{
   /*----------------------------------------------------------------------*/
   const char *co_p_c_url_base_fmt
       = "http://sivideo.webservices.francetelevisions.fr/tools/"
         "getInfosOeuvre/v2/?idDiffusion=SIM_France%s" ;
   wxString s_url = wxString::Format( co_p_c_url_base_fmt, m_s_id ) ;
   wxString s_data ;
   int i_err = sc::download( s_url, s_data ) ;
   if( i_err == sc::CDownloadBase::DE_FILE_NOT_FOUND )
   {  set_gs_error( GSERR_UNKNOWN_CHANNEL ) ; return -1 ; }
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  set_gs_error( GSERR_DOWNLOAD ) ; return -2 ; }

   /*--( We are looking for the geoblocked url ... )-----------------------*/
   const wxString co_s_re_url
       = R"=("format":"dash","url":"([^"]*)","url_secure":"[^"]*",)="
         R"=("statut":"ONLINE","drm":false,"embed":null,"geoblocage":\[)=" ;
   if(    sc::match( s_data, co_s_re_url, &s_url ) <= 1
       || s_url.empty()
     )
   {  set_gs_error( GSERR_PARSING ) ; return -3 ; }

   /*----------------------------------------------------------------------*/
   s_url.Replace( R"=(\/)=", "/", true ) ;
   s_url =   "http://hdfauth.francetv.fr/esi/TA?format=json&url="
           + mtf::url_encode( s_url ) ;
   /*----------------------------------------------------------------------*/
   i_err = sc::download( s_url, s_data ) ;
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  set_gs_error( GSERR_PARSING ) ; return -4 ; }

   /*----------------------------------------------------------------------*/
   if(    sc::match( s_data, R"=("url":\s*"([^"]*)")=", &s_url ) <= 1
       || s_url.empty()
     )
   {  set_gs_error( GSERR_PARSING ) ; return -5 ; }

   /*-----------------------------------------------------------------------+
   ! The audio stream is separated from the video ones. Therefore only the  !
   ! HLS stream is added.                                                   !
   +-----------------------------------------------------------------------*/
   add_hls_stream( s_url ) ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*--------------------------------------------------------------------------+
! The code below is based on "catchuptvandmore"'s one                       !
+--------------------------------------------------------------------------*/
int CChannelDynamicKnown::lookup_streams_mytf1()
{
   /*----------------------------------------------------------------------*/
   const char *co_p_c_url
            = "https://delivery.tf1.fr/mytf1-wrd/L_%s?format=hls" ;
   /*----------------------------------------------------------------------*/
   wxString s_url = wxString::Format( co_p_c_url, m_s_id.Upper() ) ;
   wxString s_data ;
   int i_err = sc::download( s_url, s_data, nullptr ) ;
   if( i_err == sc::CDownloadBase::DE_FILE_NOT_FOUND )
   {  set_gs_error( GSERR_UNKNOWN_CHANNEL ) ; return -1 ; }
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  set_gs_error( GSERR_DOWNLOAD ) ; return -2 ; }

   /*--( Extract hls url )-------------------------------------------------*/
   if(    sc::match( s_data, R"=("url":"(.*?)")=", &s_url ) <= 1
       || s_url.empty()
     )
   {  set_gs_error( GSERR_PARSING ) ; return -3 ; }

   /*----------------------------------------------------------------------*/
   if( hls_extract_streams( s_url ) != 0 ) { return -4 ; }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelDynamicKnown::lookup_streams_gulli()
{
   /*----------------------------------------------------------------------*/
   const char *co_p_c_url = "http://replay.gulli.fr/jwplayer/embedstreamtv" ;
   wxString s_data ;
   int i_err = sc::download( co_p_c_url, s_data ) ;
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  set_gs_error( GSERR_DOWNLOAD ) ; return -1 ; }

   /*--( Extract hls url )-------------------------------------------------*/
   wxString s_url ;
   if(    sc::match( s_data, R"=(\mfile: '(.*?)')=", &s_url ) <= 1
       || s_url.empty()
     )
   {  set_gs_error( GSERR_PARSING ) ; return -2 ; }

   /*----------------------------------------------------------------------*/
   if( hls_extract_streams( s_url ) != 0 ) { return -3 ; }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CChannelDynamicKnown::lookup_streams_melody()
{
   /*----------------------------------------------------------------------*/
   const char *co_p_c_url = "http://www.melody.tv/en-direct" ;
   wxString s_data ;
   int i_err = sc::download( co_p_c_url, s_data ) ;
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  set_gs_error( GSERR_DOWNLOAD ) ; return -1 ; }

   /*--( Extract hls url )-------------------------------------------------*/
   wxString s_url ;
   if(    sc::match( s_data, R"=(\mfile: '(.*?)')=", &s_url ) <= 1
       || s_url.empty()
     )
   {  set_gs_error( GSERR_PARSING ) ; return -2 ; }

   /*----------------------------------------------------------------------*/
   if( hls_extract_streams( s_url ) != 0 ) { return -3 ; }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*--------------------------------------------------------------------------+
! The code below is based on "catchuptvandmore"'s one                       !
+--------------------------------------------------------------------------*/
int CChannelDynamicKnown::lookup_streams_bfm()
{
   /*----------------------------------------------------------------------*/
   wxString s_url ;
   if( m_s_id == "bfmtv" )
   {  s_url = "http://www.bfmtv.com/mediaplayer/live-video/" ; }
   else
   if( m_s_id == "rmcdecouverte" )
   {  s_url = "http://rmcdecouverte.bfmtv.com/mediaplayer-direct/" ; }
   else
   if( m_s_id == "bfmbusiness" )
   {  s_url = "http://bfmbusiness.bfmtv.com/mediaplayer/live-video/" ; }
   else
   if( m_s_id == "bfmparis" )
   {  s_url = "http://www.bfmtv.com/mediaplayer/live-bfm-paris/" ; }
   else
   if( m_s_id == "bfmlyon" )
   {  s_url = "https://www.bfmtv.com/lyon/en-direct/" ; }
   else
   {  set_gs_error( GSERR_UNKNOWN_CHANNEL ) ;
      return -1 ;
   }

   /*----------------------------------------------------------------------*/
   wxString s_data ;
   int i_err = sc::download( s_url, s_data ) ;
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  set_gs_error( GSERR_DOWNLOAD ) ; return -2 ; }

   /*----------------------------------------------------------------------*/
   wxString s_data_account  ;
   wxString s_data_player   ;
   wxString s_data_video_id ;

   /*--( Extract data needed by brightcore )-------------------------------*/
   if(    sc::match( s_data, R"=(\mdata-account="(.*?)")=",
                     &s_data_account
                   ) <= 1
       || s_data_account.empty()
       || sc::match( s_data, R"=(\mdata-player="(.*?)")=",
                     &s_data_player
                   ) <= 1
       || s_data_player.empty()
       || sc::match( s_data, R"=(\mdata-video-id="(.*?)")=",
                     &s_data_video_id
                   ) <= 1
       || s_data_video_id.empty()
     )
   {  set_gs_error( GSERR_PARSING ) ; return -3 ; }

   /*--( Brightcove policy key )-------------------------------------------*/
   const char *co_p_c_url_bpk_fmt
            = "http://players.brightcove.net/%s/%s_default/index.min.js" ;
   s_url.Printf( co_p_c_url_bpk_fmt, s_data_account, s_data_player ) ;
   i_err = sc::download( s_url, s_data ) ;
   if( i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  set_gs_error( GSERR_DOWNLOAD ) ; return -4 ; }
   /*----------------------------------------------------------------------*/
   wxString s_bpk ;
   if(    sc::match( s_data, R"=(\mpolicyKey:"(.*?)")=", &s_bpk ) <= 1
       || s_bpk.empty()
     )
   {  set_gs_error( GSERR_PARSING ) ; return -5 ; }

   /*--( Json file )-------------------------------------------------------*/
   const char *co_p_c_url_bvj_fmt
     = "https://edge.api.brightcove.com/playback/v1/accounts/%s/videos/%s" ;
   s_url.Printf( co_p_c_url_bvj_fmt, s_data_account, s_data_video_id ) ;
   i_err = sc::download( s_url, s_data, nullptr, -1,
                         { "Accept: application/json;pk=" + s_bpk }
                       ) ;
   /*--( A "301" can be issued but the data is present )-------------------*/
   if( i_err == sc::CDownloadBase::DE_FORBIDDEN )
   {  set_gs_error( GSERR_FORBIDDEN ) ; return -6 ; }
   if( !s_data.empty() && i_err != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  set_gs_error( GSERR_DOWNLOAD ) ; return -7 ; }

   /*----------------------------------------------------------------------*/
   if(    sc::match( s_data, R"=("sources":.*?"src":"(.*?)")=", &s_url ) <= 1
       || s_bpk.empty()
     )
   {  set_gs_error( GSERR_PARSING ) ; return -8 ; }

   /*-----------------------------------------------------------------------+
   ! The audio stream is separated from the video ones. Therefore only the  !
   ! HLS stream is added.                                                   !
   +-----------------------------------------------------------------------*/
   add_hls_stream( s_url ) ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/



/*==========================================================================+
!                         End of file CChannel.cpp                          !
+==========================================================================*/
